import datetime
import json
import logging
import math
import mimetypes
import os
import re
import sys
import uuid

import requests

from urllib.parse import urljoin
from wsgiref.util import FileWrapper
from xml.dom import minidom, Node


from django.conf import settings
from django.core.files.storage import get_storage_class
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.validators import ValidationError
from django.db import IntegrityError
from django.http import HttpResponse, Http404
from django.http import HttpResponseNotFound, StreamingHttpResponse
from django.utils import timezone

from rest_framework import exceptions

from .tags import XFORM_ID_STRING, VERSION

PENDING = 0
SUCCESSFUL = 1
FAILED = 2

EXTERNAL_EXPORT_TYPES = ['xls']
EXPORT_EXT = {
    'csv': 'csv',
    'csvzip': 'csv_zip',
    'kml': 'kml',
    'savzip': 'sav_zip',
    'uuid': 'external',
    'xls': 'xls',
    'xlsx': 'xls',
    'zip': 'zip',
}


class XLSFormError(Exception):
    pass


class DuplicateInstance(Exception):
    def __str__(self):
        return 'Duplicate Instance'


class InstanceInvalidUserError(Exception):
    def __str__(self):
        return 'Could not determine the user.'


class InstanceParseError(Exception):
    def __str__(self):
        return 'The instance could not be parsed.'


class InstanceEmptyError(InstanceParseError):
    def __str__(self):
        return 'Empty instance'


class NonUniqueFormIdError(Exception):
    pass


class InstanceMultipleNodeError(Exception):
    pass


class FormIsMergedDatasetError(Exception):
    """Exception class for merged datasets"""

    def __str__(self):
        return 'Submissions are not allowed on merged datasets.'


class FormInactiveError(Exception):
    """Exception class for inactive forms"""

    def __str__(self):
        return 'Form is inactive'


def generate_content_disposition_header(name, extension, show_date=True):
    if name is None:
        return 'attachment;'
    if show_date:
        name = "%s-%s" % (name, timezone.now().strftime("%Y-%m-%d-%H-%M-%S"))
    return 'attachment; filename=%s.%s' % (name, extension)


def _get_all_attributes(node):
    """
    Go through an XML document returning all the attributes we see.
    """
    if hasattr(node, "hasAttributes") and node.hasAttributes():
        for key in node.attributes.keys():
            yield key, node.getAttribute(key)

    for child in node.childNodes:
        for pair in _get_all_attributes(child):
            yield pair


def _flatten_dict_nest_repeats(d, prefix):
    """
    Return a list of XPath, value pairs.

    :param d: A dictionary
    :param prefix: A list of prefixes
    """
    for key, value in d.items():
        new_prefix = prefix + [key]
        if isinstance(value, dict):
            for pair in _flatten_dict_nest_repeats(value, new_prefix):
                yield pair
        elif isinstance(value, list):
            repeats = []

            for i, item in enumerate(value):
                item_prefix = list(new_prefix)  # make a copy
                if isinstance(item, dict):
                    repeat = {}

                    for path, value in _flatten_dict_nest_repeats(
                            item, item_prefix):
                        # TODO: this only considers the first level of repeats
                        repeat.update({u"/".join(path[1:]): value})
                    repeats.append(repeat)
                else:
                    repeats.append({u"/".join(item_prefix[1:]): item})
            yield (new_prefix, repeats)
        else:
            yield (new_prefix, value)


def _gather_parent_node_list(node):
    node_names = []

    # also check for grand-parent node to skip document element
    if node.parentNode and node.parentNode.parentNode:
        node_names.extend(_gather_parent_node_list(node.parentNode))

    node_names.extend([node.nodeName])

    return node_names


def xpath_from_xml_node(node):
    node_names = _gather_parent_node_list(node)

    return "/".join(node_names[1:])


def _xml_node_to_dict(node, repeats=[], encrypted=False):
    if len(node.childNodes) == 0:
        # there's no data for this leaf node
        return None
    elif len(node.childNodes) == 1 and \
            node.childNodes[0].nodeType == node.TEXT_NODE:
        # there is data for this leaf node
        return {node.nodeName: node.childNodes[0].nodeValue}
    else:
        # this is an internal node
        value = {}

        for child in node.childNodes:
            # handle CDATA text section
            if child.nodeType == child.CDATA_SECTION_NODE:
                return {child.parentNode.nodeName: child.nodeValue}

            d = _xml_node_to_dict(child, repeats)

            if d is None:
                continue

            child_name = child.nodeName
            child_xpath = xpath_from_xml_node(child)
            if list(d) != [child_name]:
                raise AssertionError()
            node_type = dict
            # check if name is in list of repeats and make it a list if so
            # All the photo attachments in an encrypted form use name media
            if child_xpath in repeats or (encrypted and child_name == 'media'):
                node_type = list

            if node_type == dict:
                if child_name not in value:
                    value[child_name] = d[child_name]
                else:
                    # node is repeated, aggregate node values
                    node_value = value[child_name]
                    # 1. check if the node values is a list
                    if not isinstance(node_value, list):
                        # if not a list create
                        value[child_name] = [node_value]
                    # 2. parse the node
                    d = _xml_node_to_dict(child, repeats)
                    # 3. aggregate
                    value[child_name].append(d[child_name])
            else:
                if child_name not in value:
                    value[child_name] = [d[child_name]]
                else:
                    value[child_name].append(d[child_name])
        if value == {}:
            return None
        else:
            return {node.nodeName: value}


def set_uuid(obj):
    """
    Only give an object a new UUID if it does not have one.
    """
    if not obj.uuid:
        obj.uuid = uuid.uuid4().hex


def clean_and_parse_xml(xml_string):
    clean_xml_str = xml_string.strip()
    try:
        clean_xml_str = clean_xml_str.decode("utf-8")
    except Exception:
        pass
    clean_xml_str = re.sub(r">\s+<", u"><", clean_xml_str)
    xml_obj = minidom.parseString(clean_xml_str)
    return xml_obj


def get_meta_from_xml(xml_str, meta_name):
    xml = clean_and_parse_xml(xml_str)
    children = xml.childNodes
    # children ideally contains a single element
    # that is the parent of all survey elements
    if children.length == 0:
        raise ValueError("XML string must have a survey element.")
    survey_node = children[0]
    meta_tags = [n for n in survey_node.childNodes if
                 n.nodeType == Node.ELEMENT_NODE and
                 (n.tagName.lower() == "meta" or
                     n.tagName.lower() == "orx:meta")]
    if len(meta_tags) == 0:
        return None

    # get the requested tag
    meta_tag = meta_tags[0]
    uuid_tags = [n for n in meta_tag.childNodes if
                 n.nodeType == Node.ELEMENT_NODE and
                 (n.tagName.lower() == meta_name.lower() or
                  n.tagName.lower() == u'orx:%s' % meta_name.lower())]
    if len(uuid_tags) == 0:
        return None

    uuid_tag = uuid_tags[0]
    return uuid_tag.firstChild.nodeValue.strip() if uuid_tag.firstChild\
        else None


def flatten(l):
    return [item for sublist in l for item in sublist]


def _get_fields_of_type(xform, types):
    k = []
    survey_elements = flatten(
        [xform.get_survey_elements_of_type(t) for t in types])

    for element in survey_elements:
        name = element.get_abbreviated_xpath()
        k.append(name)

    return k


def get_numeric_fields(xform):
    """List of numeric field names for specified xform"""
    return _get_fields_of_type(xform, ['decimal', 'integer'])


def get_uuid_from_xml(xml):
    def _uuid_only(uuid, regex):
        matches = regex.match(uuid)
        if matches and len(matches.groups()) > 0:
            return matches.groups()[0]
        return None
    uuid = get_meta_from_xml(xml, "instanceID")
    regex = re.compile(r"uuid:(.*)")
    if uuid:
        return _uuid_only(uuid, regex)
    # check in survey_node attributes
    xml = clean_and_parse_xml(xml)
    children = xml.childNodes
    # children ideally contains a single element
    # that is the parent of all survey elements
    if children.length == 0:
        raise ValueError("XML string must have a survey element.")
    survey_node = children[0]
    uuid = survey_node.getAttribute('instanceID')
    if uuid != '':
        return _uuid_only(uuid, regex)
    return None


def numeric_checker(string_value):
    if string_value.isdigit():
        return int(string_value)
    else:
        try:
            value = float(string_value)
            if math.isnan(value):
                value = 0
            return value
        except ValueError:
            pass


def get_values_matching_key(doc, key):
    """
    Returns iterator of values in 'doc' with the matching 'key'.
    """
    def _get_values(doc, key):
        if doc is not None:
            if key in doc:
                yield doc[key]

            for z in doc.items():
                v = z[1]
                if isinstance(v, dict):
                    for item in _get_values(v, key):
                        yield item
                elif isinstance(v, list):
                    for i in v:
                        for j in _get_values(i, key):
                            yield j

    return _get_values(doc, key)


class XFormInstanceParser(object):

    def __init__(self, xml_str, data_dictionary):
        self.dd = data_dictionary
        self.parse(xml_str)

    def parse(self, xml_str):
        self._xml_obj = clean_and_parse_xml(xml_str)
        self._root_node = self._xml_obj.documentElement
        repeats = [e.get_abbreviated_xpath()
                   for e in self.dd.get_survey_elements_of_type(u"repeat")]

        self._dict = _xml_node_to_dict(self._root_node, repeats)
        self._flat_dict = {}

        if self._dict is None:
            raise InstanceEmptyError

        for path, value in _flatten_dict_nest_repeats(self._dict, []):
            self._flat_dict[u"/".join(path[1:])] = value
        self._set_attributes()

    def get_root_node(self):
        return self._root_node

    def get_root_node_name(self):
        return self._root_node.nodeName

    def get(self, abbreviated_xpath):
        return self.to_flat_dict()[abbreviated_xpath]

    def to_dict(self):
        return self._dict

    def to_flat_dict(self):
        return self._flat_dict

    def get_attributes(self):
        return self._attributes

    def _set_attributes(self):
        self._attributes = {}
        all_attributes = list(_get_all_attributes(self._root_node))
        for key, value in all_attributes:
            # Since enketo forms may have the template attribute in
            # multiple xml tags, overriding and log when this occurs
            if key in self._attributes:
                logger = logging.getLogger("console_logger")
                logger.debug("Skipping duplicate attribute: %s"
                             " with value %s" % (key, value))
                logger.debug(str(all_attributes))
            else:
                self._attributes[key] = value

    def get_xform_id_string(self):
        return self._attributes[u"id"]

    def get_version(self):
        return self._attributes.get(u"version")

    def get_flat_dict_with_attributes(self):
        result = self.to_flat_dict().copy()
        result[XFORM_ID_STRING] = self.get_xform_id_string()

        version = self.get_version()
        if version:
            result[VERSION] = self.get_version()

        return result


def response_with_mimetype_and_name(mimetype,
                                    name,
                                    extension=None,
                                    show_date=True,
                                    file_path=None,
                                    use_local_filesystem=False,
                                    full_mime=False):
    if extension is None:
        extension = mimetype
    if not full_mime:
        mimetype = "application/%s" % mimetype
    if file_path:
        try:
            if isinstance(file_path, InMemoryUploadedFile):
                response = StreamingHttpResponse(
                    file_path, content_type=mimetype)
                response['Content-Length'] = file_path.size
            elif not use_local_filesystem:
                default_storage = get_storage_class()()
                wrapper = FileWrapper(default_storage.open(file_path))
                response = StreamingHttpResponse(
                    wrapper, content_type=mimetype)
                response['Content-Length'] = default_storage.size(file_path)
            else:
                wrapper = FileWrapper(open(file_path))
                response = StreamingHttpResponse(
                    wrapper, content_type=mimetype)
                response['Content-Length'] = os.path.getsize(file_path)
        except IOError:
            response = HttpResponseNotFound(
                "The requested file could not be found.")
    else:
        response = HttpResponse(content_type=mimetype)
    response['Content-Disposition'] = generate_content_disposition_header(
        name, extension, show_date)
    return response


def _get_export_type(export_type):
    if export_type in list(EXPORT_EXT):
        export_type = EXPORT_EXT[export_type]
    else:
        raise exceptions.ParseError(
            "'%(export_type)s' format not known or not implemented!" %
            {'export_type': export_type})

    return export_type


def get_file_extension(content_type):
    return mimetypes.guess_extension(content_type)[1:]


def get_media_file_response(metadata, username=None):
    """
    Returns a HTTP response for media files.

    HttpResponse 200 if it represents a file on disk.
    HttpResponseRedirect 302 incase the metadata represents a url.
    HttpResponseNotFound 404 if the metadata file cannot be found.
    """

    if metadata.data_type == 'media' and metadata.data_file:
        file_path = metadata.data_file.name
        filename, extension = os.path.splitext(file_path.split('/')[-1])
        extension = extension.strip('.')
        dfs = get_storage_class()()
        if dfs.exists(file_path):
            return response_with_mimetype_and_name(
                metadata.data_file_type,
                filename,
                extension=extension,
                show_date=False,
                file_path=file_path,
                full_mime=True)
    elif metadata.data_type == 'url' and not metadata.data_file:
        url = requests.Request(
            'GET', metadata.data_value, params={
                'username': username
            }
        ).prepare().url

        try:
            data_file = metadata.get_file(url)
        except Exception:
            raise Http404
        return response_with_mimetype_and_name(
            mimetype=data_file.content_type,
            name=data_file.name,
            extension=get_file_extension(data_file.content_type),
            show_date=False,
            file_path=data_file,
            use_local_filesystem=False,
            full_mime=True
        )
    return HttpResponseNotFound()


def report_exception(*args, **kwargs):
    # dummy
    return


def publish_form(callback):
    """
    Calls the callback function to publish a XLSForm and returns appropriate
    message depending on exception throw during publishing of a XLSForm.
    """
    try:
        return callback()
    # except (PyXFormError, XLSFormError) as e:
    #     return {'type': 'alert-error', 'text': str(e)}
    except IntegrityError as e:
        return {
            'type': 'alert-error',
            'text': 'Form with this id or SMS-keyword already exists.',
        }
    # except ProcessTimedOut as e:
    #     # catch timeout errors
    #     return {
    #         'type': 'alert-error',
    #         'text': 'Form validation timeout, please try again.',
    #     }
    except (MemoryError, OSError) as e:
        return {
            'type': 'alert-error',
            'text': (
                'An error occurred while publishing the form. '
                'Please try again.'
            ),
        }
    except (AttributeError, Exception, ValidationError) as e:
        report_exception("Form publishing exception: {}".format(e), str(e),
                         sys.exc_info())
        return {'type': 'alert-error', 'text': str(e)}


def _get_tag_or_element_type_xpath(xform, tag):
    elems = xform.get_survey_elements_of_type(tag)

    return elems[0].get_abbreviated_xpath() if elems else tag


def calculate_duration(start_time, end_time):
    """
    This function calculates duration when given start and end times.
    An empty string is returned if either of the time formats does
    not match '_format' format else, the duration is returned
    """
    _format = "%Y-%m-%dT%H:%M:%S"
    try:
        _start = datetime.datetime.strptime(start_time[:19], _format)
        _end = datetime.datetime.strptime(end_time[:19], _format)
    except (TypeError, ValueError):
        return ''

    duration = (_end - _start).total_seconds()

    return duration


def inject_instanceid(xml_str, uuid):
    if get_uuid_from_xml(xml_str) is None:
        xml = clean_and_parse_xml(xml_str)
        children = xml.childNodes
        if children.length == 0:
            raise ValueError("XML string must have a survey element.")

        # check if we have a meta tag
        survey_node = children.item(0)
        meta_tags = [
            n for n in survey_node.childNodes
            if n.nodeType == Node.ELEMENT_NODE and n.tagName.lower() == "meta"
        ]
        if len(meta_tags) == 0:
            meta_tag = xml.createElement("meta")
            xml.documentElement.appendChild(meta_tag)
        else:
            meta_tag = meta_tags[0]

        # check if we have an instanceID tag
        uuid_tags = [
            n for n in meta_tag.childNodes
            if n.nodeType == Node.ELEMENT_NODE and n.tagName == "instanceID"
        ]
        if len(uuid_tags) == 0:
            uuid_tag = xml.createElement("instanceID")
            meta_tag.appendChild(uuid_tag)
        else:
            uuid_tag = uuid_tags[0]
        # insert meta and instanceID
        text_node = xml.createTextNode(u"uuid:%s" % uuid)
        uuid_tag.appendChild(text_node)
        return xml.toxml()
    return xml_str


class EnketoError(Exception):

    default_message = "There was a problem with your submissionor form. Please contact support."

    def __init__(self, message=None):
        if message is None:
            self.message = self.default_message
        else:
            self.message = message

    def __str__(self):
        return "{}".format(self.message)


def handle_enketo_error(response):
    """Handle enketo error response."""
    try:
        data = json.loads(response.content)
    except ValueError:
        pass

        if response.status_code == 502:
            raise EnketoError(
                u"Sorry, we cannot load your form right now.  Please try "
                "again later.")
        raise EnketoError()
    else:
        if 'message' in data:
            raise EnketoError(data['message'])
        raise EnketoError(response.text)


def enketo_url(
    form_url, id_string, instance_xml=None,
    instance_id=None,
    return_url=None,
    offline=False
):
    if (not hasattr(settings, 'ENKETO_URL') or
            not hasattr(settings, 'ENKETO_API_SURVEY_PATH') or
            not hasattr(settings, 'ENKETO_API_TOKEN') or
            settings.ENKETO_API_TOKEN == ''):
        return False

    values = {'form_id': id_string, 'server_url': form_url}
    if instance_id and instance_xml:
        url = urljoin(settings.ENKETO_URL, settings.ENKETO_API_INSTANCE_PATH)
        values.update({
            'instance': instance_xml,
            'instance_id': instance_id,
            'return_url': return_url
        })
    else:
        survey_path = settings.ENKETO_API_SURVEY_PATH
        if offline:
            survey_path += '/offline'
        url = urljoin(settings.ENKETO_URL, survey_path)

    response = requests.post(
        url,
        data=values,
        auth=(settings.ENKETO_API_TOKEN, ''),
        verify=getattr(settings, 'ENKETO_VERIFY_SSL', False))
    if response.status_code in (200, 201):
        try:
            data = json.loads(response.content)
        except ValueError:
            pass
        else:
            url = (data.get('edit_url') or data.get('offline_url') or
                   data.get('url'))
            if url:
                return url

    handle_enketo_error(response)


def get_form_url(
    request, protocol='http', preview=False,  # xform_pk=None
):
    """
    Return a form list url endpoint to be used to make a request to Enketo.

    For example, it will return https://example.com and Enketo will know to
    look for the form list at https://example.com/formList. If a username is
    provided then Enketo will request the form list from
    https://example.com/[username]/formList. Same applies for preview if
    preview is True and also to a single form when xform_pk is provided.
    """
    http_host = request.META.get('HTTP_HOST', 'dev.monitora.sisicmbio.icmbio.gov.br')
    url = '%s://%s' % (protocol, http_host)
    if preview:
        url = '%s/preview' % url
    return "{}/xform".format(url)



ATTACHMENTS = "_attachments"
DURATION = "_duration"
EDITED = "_edited"
END = "end"
END_TIME = "end_time"
GEOLOCATION = "_geolocation"
GROUP_DELIMETER_TAG = 'group_delimiter'
ID = "_id"
KNOWN_MEDIA_TYPES = ['photo', 'image', 'audio', 'video']
LAST_EDITED = "_last_edited"
MEDIA_ALL_RECEIVED = "_media_all_received"
MEDIA_COUNT = "_media_count"
NOTES = "_notes"
REPEAT_INDEX_TAGS = 'repeat_index_tags'
START = "start"  # This tag was used in Phase II
START_TIME = "start_time"  # This tag was used in Phase I
STATUS = "_status"
SUBMISSION_TIME = '_submission_time'
SUBMITTED_BY = "_submitted_by"
TOTAL_MEDIA = "_total_media"
UUID = "_uuid"
VERSION = "_version"
XFORM_ID = "_xform_id"
XFORM_ID_STRING = "_xform_id_string"



from django import forms
from django.db import transaction

from .models import DataDictionary


class QuickConverterFile(forms.Form):
    """
    Uploads XLSForm form.
    """
    xls_file = forms.FileField(
        label='XLS File', required=False)


@transaction.atomic()
def publish_xls_form(xls_file, user, id_string=None, created_by=None):
    """Create or update DataDictionary with xls_file, user
    id_string is optional when updating
    """
    # get or create DataDictionary based on user and id string
    if id_string:
        dd = DataDictionary.objects.get(id_string=id_string)
        dd.xls = xls_file
        dd.save()

        return dd
    else:
        return DataDictionary.objects.create(
            created_by=created_by or user,
            user=user,
            xls=xls_file,
        )


class QuickConverter(QuickConverterFile):
    """
    Publish XLSForm and convert to XForm.
    """

    def publish(self, user, id_string=None, created_by=None):
        """
        Publish XLSForm.
        """
        if self.is_valid():
            # If a text (csv) representation of the xlsform is present,
            # this will save the file and pass it instead of the 'xls_file'
            # field.
            cleaned_xls_file = None

            if 'xls_file' in self.cleaned_data and\
                    self.cleaned_data['xls_file']:
                cleaned_xls_file = self.cleaned_data['xls_file']

            # publish the xls
            return publish_xls_form(
                cleaned_xls_file, user,
                id_string, created_by or user)



from django.urls import re_path

from xform import views
from xform import enketo_views


urlpatterns = [
    re_path(
        r'^(?P<id_string>[^/]+)/enter_data/$',
        enketo_views.enter_data,
        name='enter_data'
    ),
    re_path(
        r'^(?P<id_string>[^/]+)/enter_data/offline/$',
        enketo_views.enter_data,
        {'offline': True},
        name='enter_data_offline'
    ),
    re_path(
        r'^(?P<id_string>[^/]+)/edit_data/(?P<data_id>\d+)/$',
        enketo_views.edit_data,
        name='edit_data'
    ),
    re_path(
        r'^(?P<username>[^/]+)/$',
        views.profile,
        name='user_profile'
    ),
    re_path(
        r'^formList$',
        views.XFormListViewSet.as_view({'get': 'list', 'head': 'list'}),
        name='form-list'),
    re_path(
        r'^(?P<username>\w+)/xformsManifest/(?P<pk>[\d+^/]+)$',
        views.XFormListViewSet.as_view({'get': 'manifest', 'head': 'manifest'}),
        name='manifest-url'),
    re_path(
        r'^(?P<username>\w+)/forms/(?P<pk>[\d+^/]+)/form\.xml$',
        views.XFormListViewSet.as_view({'get': 'retrieve', 'head': 'retrieve'}),
        name='download_xform'),
    re_path(
        r'^xformsMedia/(?P<pk>[\d+^/]+)/(?P<metadata>[\d+^/.]+)\.(?P<format>([a-z]|[0-9])*)$',
        views.XFormListViewSet.as_view({'get': 'media', 'head': 'media'}),
        name='xform-media'),
    re_path(
        r'^(?P<username>\w+)/xformsMedia/(?P<pk>[\d+^/]+)/(?P<metadata>[\d+^/.]+)$',
        views.XFormListViewSet.as_view({'get': 'media', 'head': 'media'}),
        name='xform-media'),
    re_path(
        r'^(?P<username>\w+)/xformsMedia/(?P<pk>[\d+^/]+)/(?P<metadata>[\d+^/.]+)\.(?P<format>([a-z]|[0-9])*)$',
        views.XFormListViewSet.as_view({'get': 'media', 'head': 'media'}),
        name='xform-media'),
    re_path(
        r'^xformsMedia/(?P<pk>[\d+^/]+)/(?P<metadata>[\d+^/.]+)$',
        views.XFormListViewSet.as_view({'get': 'media', 'head': 'media'}),
        name='xform-media'),
    re_path(
        r'^submission$',
        views.XFormSubmissionView.as_view(),
        name='submissions'),
    re_path(
        r'^(?P<username>\w+)/(?P<id_string>[^/]+)/edit/$',
        views.edit,
        name='xform_edit'),
]



import csv
import json
import mimetypes
import os
import random
import re

import requests
import xlrd

from contextlib import closing
from hashlib import md5
from io import BytesIO
from io import StringIO


from pyxform import SurveyElementBuilder
from pyxform.builder import create_survey_element_from_dict
from pyxform.utils import has_external_choices
from pyxform.xform2json import create_survey_element_from_xml
from pyxform.xls2json import parse_file_to_json

from xml.dom import Node

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.contrib.gis.db import models
from django.contrib.gis.geos import GeometryCollection, Point

from django.core.exceptions import ValidationError
from django.core.files.temp import NamedTemporaryFile
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.validators import URLValidator
from django.db.models.signals import post_save
from django.utils import timezone


from .tags import (
    UUID, ID, ATTACHMENTS, STATUS, NOTES, VERSION, DURATION, XFORM_ID_STRING,
    XFORM_ID, GEOLOCATION, SUBMITTED_BY, SUBMISSION_TIME, TOTAL_MEDIA,
    MEDIA_COUNT, MEDIA_ALL_RECEIVED, EDITED, LAST_EDITED, KNOWN_MEDIA_TYPES,
    START, END
)
from .utils import (
    get_values_matching_key, get_uuid_from_xml, set_uuid, XFormInstanceParser,
    clean_and_parse_xml, get_numeric_fields, numeric_checker,
    _get_tag_or_element_type_xpath, calculate_duration
)

if 'mysql' in settings.DATABASES['default']['ENGINE']:
    from jsonfield import JSONField
else:
    from django.contrib.postgres.fields import JSONField

CHUNK_SIZE = 1024
XFORM_TITLE_LENGTH = 255
title_pattern = re.compile(r"<h:title>(.*?)</h:title>")


def contains_xml_invalid_char(text, invalids=['&', '>', '<']):
    """Check whether 'text' contains ANY invalid xml chars"""
    return 1 in [c in text for c in invalids]


def convert_to_serializable_date(date):
    if hasattr(date, 'isoformat'):
        return date.isoformat()
    return date


def _get_attachments_from_instance(instance):
    attachments = []
    for a in instance.attachments.all():
        attachment = dict()
        attachment['download_url'] = a.media_file.url
        attachment['small_download_url'] = a.media_file.url
        attachment['medium_download_url'] = a.media_file.url
        attachment['mimetype'] = a.mimetype
        attachment['filename'] = a.media_file.name
        attachment['name'] = a.name
        attachment['instance'] = a.instance.pk
        attachment['xform'] = instance.xform.id
        attachment['id'] = a.id
        attachments.append(attachment)

    return attachments


def get_default_content_type():
    content_object, created = ContentType.objects.get_or_create(
        app_label="xform", model='xform')
    return content_object.id


def upload_to(instance, filename):
    try:
        return os.path.join(
            instance.user.username, 'xls',
            os.path.split(filename)[1])
    except Exception:
        folder = "{}_{}".format(instance.instance.xform.id,
                                instance.instance.xform.id_string)
        return os.path.join(
            instance.instance.xform.user.username, 'attachments', folder,
            os.path.split(filename)[1])


class XLSFormError(Exception):
    pass


class FormInactiveError(Exception):
    pass


class XForm(models.Model):
    dynamic_choices = True

    xls = models.FileField(upload_to=upload_to, null=True)
    json = models.TextField(default=u'')
    description = models.TextField(default=u'', null=True, blank=True)
    xml = models.TextField()

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='xforms', null=True, on_delete=models.CASCADE)

    id_string = models.SlugField(
        editable=False,
        verbose_name="ID",
        max_length=100)
    title = models.CharField(editable=False, max_length=255)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)

    last_submission_time = models.DateTimeField(blank=True, null=True)
    has_start_time = models.BooleanField(default=False)
    uuid = models.CharField(max_length=36, default=u'')

    uuid_regex = re.compile(r'(<instance>.*?id="[^"]+">)(.*</instance>)(.*)',
                            re.DOTALL)
    instance_id_regex = re.compile(r'<instance>.*?id="([^"]+)".*</instance>',
                                   re.DOTALL)

    instances_with_geopoints = models.BooleanField(default=False)

    num_of_submissions = models.IntegerField(default=0)

    version = models.CharField(
        max_length=255, null=True, blank=True)

    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE)

    metadata_set = GenericRelation(
        'MetaData',
        content_type_field='content_type_id',
        object_id_field="object_id")

    has_hxl_support = models.BooleanField(default=False)
    last_updated_at = models.DateTimeField(auto_now=True)
    hash = models.CharField("Hash", max_length=36, blank=True, null=True,
                            default=None)

    class Meta:
        unique_together = ("user", "id_string",)
        verbose_name = "XForm"
        verbose_name_plural = "XForms"
        ordering = ("pk", )

    def get_osm_survey_xpaths(self):
        """
        Returns abbreviated_xpath for OSM question types in the survey.
        """
        return [
            elem.get_abbreviated_xpath()
            for elem in self.get_survey_elements_of_type('osm')]

    def get_media_survey_xpaths(self):
        return [
            e.get_abbreviated_xpath()
            for e in sum([
                self.get_survey_elements_of_type(m) for m in KNOWN_MEDIA_TYPES
            ], [])
        ]

    def file_name(self):
        return self.id_string + ".xml"

    def get_survey_elements_of_type(self, element_type):
        return [
            e for e in self.get_survey_elements() if e.type == element_type
        ]

    def _set_uuid_in_xml(self, file_name=None):
        """
        Add bind to automatically set UUID node in XML.
        """
        if not file_name:
            file_name = self.file_name()
        file_name, file_ext = os.path.splitext(file_name)
        doc = clean_and_parse_xml(self.xml)
        model_nodes = doc.getElementsByTagName("model")
        if len(model_nodes) != 1:
            raise Exception(u"xml contains multiple model nodes")

        model_node = model_nodes[0]
        instance_nodes = [
            node for node in model_node.childNodes
            if node.nodeType == Node.ELEMENT_NODE and
            node.tagName.lower() == "instance" and not node.hasAttribute("id")
        ]

        if len(instance_nodes) != 1:
            raise Exception("Multiple instance nodes without the id "
                            "attribute, can't tell which is the main one")

        instance_node = instance_nodes[0]

        # get the first child whose id attribute matches our id_string
        survey_nodes = [
            node for node in instance_node.childNodes
            if node.nodeType == Node.ELEMENT_NODE and
            (node.tagName == file_name or node.attributes.get('id'))
        ]

        if len(survey_nodes) != 1:
            raise Exception(
                "Multiple survey nodes with the id '%s'" % self.id_string)

        survey_node = survey_nodes[0]
        formhub_nodes = [
            n for n in survey_node.childNodes
            if n.nodeType == Node.ELEMENT_NODE and n.tagName == "formhub"
        ]

        if len(formhub_nodes) > 1:
            raise Exception(
                "Multiple formhub nodes within main instance node")
        elif len(formhub_nodes) == 1:
            formhub_node = formhub_nodes[0]
        else:
            formhub_node = survey_node.insertBefore(
                doc.createElement("formhub"), survey_node.firstChild)

        uuid_nodes = [
            node for node in formhub_node.childNodes
            if node.nodeType == Node.ELEMENT_NODE and node.tagName == "uuid"
        ]

        if len(uuid_nodes) == 0:
            formhub_node.appendChild(doc.createElement("uuid"))
        if len(formhub_nodes) == 0:
            # append the calculate bind node
            calculate_node = doc.createElement("bind")
            calculate_node.setAttribute(
                "nodeset", "/%s/formhub/uuid" % survey_node.tagName)
            calculate_node.setAttribute("type", "string")
            calculate_node.setAttribute("calculate", "'%s'" % self.uuid)
            model_node.appendChild(calculate_node)

        self.xml = doc.toprettyxml(indent="  ", encoding='utf-8')
        # hack
        # http://ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-\
        # and-silly-whitespace/
        text_re = re.compile('(>)\n\s*(\s[^<>\s].*?)\n\s*(\s</)', re.DOTALL)
        output_re = re.compile('\n.*(<output.*>)\n(  )*')
        pretty_xml = text_re.sub(lambda m: ''.join(m.group(1, 2, 3)),
                                 self.xml.decode('utf-8'))
        inline_output = output_re.sub('\g<1>', pretty_xml)
        inline_output = re.compile('<label>\s*\n*\s*\n*\s*</label>').sub(
            '<label></label>', inline_output)
        self.xml = inline_output

    def _mark_start_time_boolean(self):
        starttime_substring = 'jr:preloadParams="start"'
        if self.xml.find(starttime_substring) != -1:
            self.has_start_time = True
        else:
            self.has_start_time = False

    def _id_string_already_exists_in_account(self, id_string):
        try:
            XForm.objects.get(user=self.user, id_string__iexact=id_string)
        except XForm.DoesNotExist:
            return False

        return True

    def get_unique_id_string(self, id_string, count=0):
        # used to generate a new id_string for new data_dictionary object if
        # id_string already existed
        if self._id_string_already_exists_in_account(id_string):
            if count != 0:
                if re.match(r'\w+_\d+$', id_string):
                    a = id_string.split('_')
                    id_string = "_".join(a[:-1])
            count += 1
            id_string = "{}_{}".format(id_string, count)

            return self.get_unique_id_string(id_string, count)

        return id_string

    def _set_title(self):
        xml = re.sub(r"\s+", " ", self.xml)
        matches = title_pattern.findall(xml)

        if len(matches) != 1:
            raise XLSFormError(("There should be a single title."), matches)

        if matches:
            title_xml = matches[0][:XFORM_TITLE_LENGTH]
        else:
            title_xml = self.title[:XFORM_TITLE_LENGTH] if self.title else ''

        if self.title and title_xml != self.title:
            title_xml = self.title[:XFORM_TITLE_LENGTH]
            if isinstance(self.xml, bytes):
                self.xml = self.xml.decode('utf-8')
            self.xml = title_pattern.sub(u"<h:title>%s</h:title>" % title_xml,
                                         self.xml)
            self._set_hash()
        if contains_xml_invalid_char(title_xml):
            raise XLSFormError(
                "Title shouldn't have any invalid xml "
                "characters ('>' '&' '<')"
            )

        self.title = title_xml

    def get_hash(self):
        return u'md5:%s' % md5(self.xml.encode('utf8')).hexdigest()

    def get_random_hash(self):
        return u'md5:%s' % md5(
            ("%s-%s" % (
                self.xml,
                random.randint(0, 25101991)
            )).encode('utf8')
        ).hexdigest()

    @property
    def random_hash(self):
        return self.get_random_hash()

    def _set_hash(self):
        self.hash = self.get_hash()

    def _set_id_string(self):
        matches = self.instance_id_regex.findall(self.xml)
        if len(matches) != 1:
            raise XLSFormError("There should be a single id string.")
        self.id_string = matches[0]

    def save(self, *args, **kwargs):
        update_fields = kwargs.get('update_fields')
        if update_fields:
            kwargs['update_fields'] = list(
                set(list(update_fields) + ['date_modified']))
        if update_fields is None or 'title' in update_fields:
            self._set_title()
        if self.pk is None:
            self._set_hash()
        if update_fields is None or 'id_string' in update_fields:
            old_id_string = self.id_string
            self._set_id_string()
            # check if we have an existing id_string,
            # if so, the one must match but only if xform is NOT new
            if self.pk and old_id_string and old_id_string != self.id_string \
                    and self.num_of_submissions > 0:
                raise XLSFormError(
                    "Your updated form's id_string '%(new_id)s' must match "
                    "the existing forms' id_string '%(old_id)s'." % {
                        'new_id': self.id_string,
                        'old_id': old_id_string
                    })
            if getattr(settings, 'STRICT', True) and \
                    not re.search(r"^[\w-]+$", self.id_string):
                raise XLSFormError(
                    'In strict mode, the XForm ID must be a '
                    'valid slug and contain no spaces.')

        if 'skip_xls_read' in kwargs:
            del kwargs['skip_xls_read']

        super(XForm, self).save(*args, **kwargs)

    def get_survey(self):
        if not hasattr(self, "_survey"):
            try:
                builder = SurveyElementBuilder()
                self._survey = \
                    builder.create_survey_element_from_json(self.json)
            except ValueError:
                xml = bytes(bytearray(self.xml, encoding='utf-8'))
                self._survey = create_survey_element_from_xml(xml)
        return self._survey

    survey = property(get_survey)

    def get_survey_elements(self):
        return self.survey.iter_descendants()

    def geopoint_xpaths(self):
        survey_elements = self.get_survey_elements()

        return [
            e.get_abbreviated_xpath() for e in survey_elements
            if e.bind.get(u'type') == u'geopoint'
        ]

    def __str__(self):
        return self.id_string


def type_for_form(content_object, data_type):
    content_type = ContentType.objects.get_for_model(content_object)
    return MetaData.objects.filter(object_id=content_object.id,
                                   content_type=content_type,
                                   data_type=data_type)


def is_valid_url(uri):
    try:
        URLValidator(uri)
    except ValidationError:
        return False

    return True


def create_media(media):
    """Download media link"""
    if is_valid_url(media.data_value):
        filename = media.data_value.split('/')[-1]
        data_file = NamedTemporaryFile()
        content_type = mimetypes.guess_type(filename)
        with closing(requests.get(media.data_value, stream=True)) as r:
            for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                if chunk:
                    data_file.write(chunk)
        data_file.seek(os.SEEK_SET, os.SEEK_END)
        size = os.path.getsize(data_file.name)
        data_file.seek(os.SEEK_SET)
        media.data_value = filename
        media.data_file = InMemoryUploadedFile(
            data_file, 'data_file', filename, content_type,
            size, charset=None)

        return media

    return None


def media_resources(media_list, download=False):
    """List of MetaData objects of type media

    @param media_list - list of MetaData objects of type `media`
    @param download - boolean, when True downloads media files when
    media.data_value is a valid url

    return a list of MetaData objects

    """
    data = []
    for media in media_list:
        if media.data_file.name == '' and download:
            media = create_media(media)

            if media:
                data.append(media)
        else:
            data.append(media)

    return data


def meta_data_upload_to(instance, filename):
    username = None

    if instance.content_object.user is None and \
            instance.content_type.model == 'instance':
        username = instance.content_object.xform.user.username
    else:
        username = instance.content_object.user.username

    if instance.data_type == 'media':
        return os.path.join(username, 'formid-media', filename)

    return os.path.join(username, 'docs', filename)


class MetaData(models.Model):
    data_type = models.CharField(max_length=255)
    data_value = models.CharField(max_length=255)
    data_file = models.FileField(upload_to=meta_data_upload_to, blank=True, null=True)
    data_file_type = models.CharField(max_length=255, blank=True, null=True)
    file_hash = models.CharField(max_length=50, blank=True, null=True)
    date_created = models.DateTimeField(null=True, auto_now_add=True)
    date_modified = models.DateTimeField(null=True, auto_now=True)
    deleted_at = models.DateTimeField(null=True, default=None)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE,
                                     default=get_default_content_type)
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    objects = models.Manager()

    class Meta:
        unique_together = ('object_id', 'data_type', 'data_value',
                           'content_type')

    def __str__(self):
        return self.data_value

    def file(self, username=None):
        if hasattr(self, '_file'):
            return self._file

        url = requests.Request(
            'GET', self.data_value, params={
                'username': username
            }
        ).prepare().url
        self._file = MetaData.get_file(url)
        return self._file

    @staticmethod
    def media_upload(content_object, data_file=None, download=False):
        data_type = 'media'
        if data_file:
            allowed_types = settings.XFORM_SUPPORTED_MEDIA_UPLOAD_TYPES
            data_content_type = data_file.content_type \
                if data_file.content_type in allowed_types else \
                mimetypes.guess_type(data_file.name)[0]

            if data_content_type in allowed_types:
                content_type = ContentType.objects.get_for_model(
                    content_object)

                media, created = MetaData.objects.update_or_create(
                    data_type=data_type,
                    content_type=content_type,
                    object_id=content_object.id,
                    data_value=data_file.name,
                    defaults={
                        'data_file': data_file,
                        'data_file_type': data_content_type
                    })
        return media_resources(
            type_for_form(content_object, data_type), download)

    @staticmethod
    def get_md5(data_file):
        hash_md5 = md5()
        for chunk in iter(lambda: data_file.read(4096), b""):
            hash_md5.update(chunk)
        return 'md5:%s' % hash_md5.hexdigest()

    @staticmethod
    def get_file(url):
        data_file = None
        output = BytesIO()

        def getsize(f):
            f.seek(0)
            f.read()
            s = f.tell()
            f.seek(0)
            return s

        r = requests.get(url, allow_redirects=True)
        d = r.headers['content-disposition']
        fname = re.findall("filename=\"(.+)\"", d)[0]
        content_type = r.headers.get('content-type')
        output.write(r.content)
        size = getsize(output)
        data_file = InMemoryUploadedFile(
            file=output, name=fname,
            field_name=None,
            content_type=content_type,
            charset='utf-8', size=size
        )
        return data_file

    @staticmethod
    def add_url(content_object, url=None, download=False):
        data_type = 'url'
        try:
            data_file = MetaData.get_file(url)
        except Exception:
            return None

        allowed_types = settings.XFORM_SUPPORTED_MEDIA_UPLOAD_TYPES
        data_content_type = data_file.content_type \
            if data_file.content_type in allowed_types else \
            mimetypes.guess_type(data_file.name)[0]

        if data_content_type in allowed_types:
            content_type = ContentType.objects.get_for_model(
                content_object)

            media, created = MetaData.objects.update_or_create(
                data_type=data_type,
                content_type=content_type,
                object_id=content_object.id,
                data_value=url,
                defaults={
                    'data_file': None,
                    'data_file_type': data_content_type
                })
        return media_resources(
            type_for_form(content_object, data_type), download)

    def save(self, *args, **kwargs):
        self._set_hash()
        super(MetaData, self).save(*args, **kwargs)

    @property
    def hash(self):
        if self.file_hash is not None and self.file_hash != '':
            return self.file_hash
        else:
            return self._set_hash()

    def _set_hash(self):
        if not self.data_file:
            return None

        file_exists = self.data_file.storage.exists(self.data_file.name)

        if (file_exists and self.data_file.name != '') \
                or (not file_exists and self.data_file):
            try:
                self.data_file.seek(os.SEEK_SET)
            except IOError:
                return ''
            else:
                self.file_hash = 'md5:%s' % md5(
                    self.data_file.read()).hexdigest()

                return self.file_hash

        return ''


class Instance(models.Model):
    """
    Model representing a single submission to an XForm
    """

    json = JSONField(default=dict, null=False)
    xml = models.TextField()
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='instances', null=True, on_delete=models.CASCADE)
    xform = models.ForeignKey('xform.XForm', null=False,
                              related_name='instances', on_delete=models.CASCADE)

    # shows when we first received this instance
    date_created = models.DateTimeField(auto_now_add=True)

    # this will end up representing "date last parsed"
    date_modified = models.DateTimeField(auto_now=True)

    # this will be edited when we need to create a new InstanceHistory object
    last_edited = models.DateTimeField(null=True, default=None)

    # ODK keeps track of three statuses for an instance:
    # incomplete, submitted, complete
    # we add a fourth status: submitted_via_web
    status = models.CharField(max_length=20,
                              default=u'submitted_via_web')
    uuid = models.CharField(max_length=249, default=u'', db_index=True)
    version = models.CharField(max_length=255, null=True)

    # store a geographic objects associated with this instance
    geom = models.GeometryCollectionField(null=True)

    # Keep track of whether all media attachments have been received
    media_all_received = models.NullBooleanField(
        "Received All Media Attachemts",
        null=True,
        default=True)
    total_media = models.PositiveIntegerField("Total Media Attachments",
                                              null=True,
                                              default=0)
    media_count = models.PositiveIntegerField("Received Media Attachments",
                                              null=True,
                                              default=0)
    checksum = models.CharField(max_length=64, null=True, blank=True,
                                db_index=True)

    class Meta:
        unique_together = ('xform', 'uuid')

    def __str__(self):
        return "Status: %s" % self.status

    @property
    def point(self):
        gc = self.geom

        if gc and len(gc):
            return gc[0]

    def get_duration(self):
        data = self.get_dict()
        # pylint: disable=no-member
        start_name = _get_tag_or_element_type_xpath(self.xform, START)
        end_name = _get_tag_or_element_type_xpath(self.xform, END)
        start_time, end_time = data.get(start_name), data.get(end_name)

        return calculate_duration(start_time, end_time)

    @property
    def num_of_media(self):
        """
        Returns number of media attachments expected in the submission.
        """
        if not hasattr(self, '_num_of_media'):
            # pylint: disable=attribute-defined-outside-init
            self._num_of_media = len(self.get_expected_media())

        return self._num_of_media

    @property
    def attachments_count(self):
        return self.attachments.filter(
            name__in=self.get_expected_media()
        ).distinct('name').order_by('name').count()

    def get_expected_media(self):
        """
        Returns a list of expected media files from the submission data.
        """
        if not hasattr(self, '_expected_media'):
            # pylint: disable=no-member
            data = self.get_dict()
            media_list = []
            if 'encryptedXmlFile' in data and self.xform.encrypted:
                media_list.append(data['encryptedXmlFile'])
                if 'media' in data:
                    # pylint: disable=no-member
                    media_list.extend([i['media/file'] for i in data['media']])
            else:
                media_xpaths = (self.xform.get_media_survey_xpaths() +
                                self.xform.get_osm_survey_xpaths())
                for media_xpath in media_xpaths:
                    media_list.extend(
                        get_values_matching_key(data, media_xpath))
            # pylint: disable=attribute-defined-outside-init
            self._expected_media = list(set(media_list))

        return self._expected_media

    def numeric_converter(self, json_dict, numeric_fields=None):
        if numeric_fields is None:
            # pylint: disable=no-member
            numeric_fields = get_numeric_fields(self.xform)
        for key, value in json_dict.items():
            if isinstance(value, (str, bytes)) and key in numeric_fields:
                converted_value = numeric_checker(value)
                if converted_value:
                    json_dict[key] = converted_value
            elif isinstance(value, dict):
                json_dict[key] = self.numeric_converter(
                    value, numeric_fields)
            elif isinstance(value, list):
                for k, v in enumerate(value):
                    if isinstance(v, (str, bytes)) and key in numeric_fields:
                        converted_value = numeric_checker(v)
                        if converted_value:
                            json_dict[key] = converted_value
                    elif isinstance(v, dict):
                        value[k] = self.numeric_converter(
                            v, numeric_fields)
        return json_dict

    def _set_geom(self):
        # pylint: disable=no-member
        xform = self.xform
        geo_xpaths = xform.geopoint_xpaths()
        doc = self.get_dict()
        points = []

        if geo_xpaths:
            for xpath in geo_xpaths:
                for gps in get_values_matching_key(doc, xpath):
                    try:
                        geometry = [float(s) for s in gps.split()]
                        lat, lng = geometry[0:2]
                        points.append(Point(lng, lat))
                    except ValueError:
                        return

            if not xform.instances_with_geopoints and len(points):
                xform.instances_with_geopoints = True
                xform.save()

            self.geom = GeometryCollection(points)

    def _check_active(self, force):
        """Check that form is active and raise exception if not.

        :param force: Ignore restrictions on saving.
        """
        # pylint: disable=no-member
        # if not force and self.xform and not self.xform.downloadable:
        #     raise FormInactiveError()
        pass

    def _set_json(self):
        self.json = self.get_full_dict()

    def get_full_dict(self, load_existing=True):
        doc = self.json or {} if load_existing else {}
        # Get latest dict
        doc = self.get_dict()
        # pylint: disable=no-member
        if self.id:
            doc.update({
                UUID: self.uuid,
                ID: self.id,
                # BAMBOO_DATASET_ID: self.xform.bamboo_dataset,
                ATTACHMENTS: _get_attachments_from_instance(self),
                STATUS: self.status,
                # TAGS: list(self.tags.names()),
                NOTES: [],
                VERSION: self.version,
                DURATION: self.get_duration(),
                XFORM_ID_STRING: self._parser.get_xform_id_string(),
                XFORM_ID: self.xform.pk,
                GEOLOCATION: [self.point.y, self.point.x] if self.point
                else [None, None],
                SUBMITTED_BY: self.user.username if self.user else None
            })

            # for osm in self.osm_data.all():
            #     doc.update(osm.get_tags_with_prefix())
            if not self.date_created:
                self.date_created = timezone.now()

            doc[SUBMISSION_TIME] = self.date_created.strftime('%Y-%m-%dT%H:%M:%S')

            doc[TOTAL_MEDIA] = self.total_media
            doc[MEDIA_COUNT] = self.media_count
            doc[MEDIA_ALL_RECEIVED] = self.media_all_received

            edited = False
            if hasattr(self, 'last_edited'):
                edited = self.last_edited is not None

            doc[EDITED] = edited
            edited and doc.update({
                LAST_EDITED: convert_to_serializable_date(self.last_edited)
            })
        return doc

    def get_dict(self, force_new=False, flat=True):
        """Return a python object representation of this instance's XML."""
        self._set_parser()

        instance_dict = self._parser.get_flat_dict_with_attributes() if flat \
            else self._parser.to_dict()
        return self.numeric_converter(instance_dict)

    def _set_survey_type(self):
        self.survey_type = self.get_root_node_name()

    def _set_parser(self):
        if not hasattr(self, "_parser"):
            # pylint: disable=no-member
            self._parser = XFormInstanceParser(self.xml, self.xform)

    def get_root_node_name(self):
        self._set_parser()
        return self._parser.get_root_node_name()

    def _set_uuid(self):
        # pylint: disable=no-member, attribute-defined-outside-init
        if self.xml and not self.uuid:
            # pylint: disable=no-member
            uuid = get_uuid_from_xml(self.xml)
            if uuid is not None:
                self.uuid = uuid
        set_uuid(self)

    def save(self, *args, **kwargs):
        force = kwargs.get('force')

        if force:
            del kwargs['force']

        # self._check_is_merged_dataset()
        self._check_active(force)
        self._set_geom()
        self._set_json()
        self._set_survey_type()
        self._set_uuid()
        # pylint: disable=no-member
        self.version = self.json.get(VERSION, self.xform.version)

        super(Instance, self).save(*args, **kwargs)


class Attachment(models.Model):
    OSM = 'osm'

    instance = models.ForeignKey(
        Instance, related_name="attachments", on_delete=models.CASCADE)
    media_file = models.FileField(
        max_length=255, upload_to=upload_to)
    mimetype = models.CharField(
        max_length=100, null=False, blank=True, default='')
    extension = models.CharField(
        max_length=10, null=False, blank=False, default=u"non", db_index=True)
    date_created = models.DateTimeField(null=True, auto_now_add=True)
    date_modified = models.DateTimeField(null=True, auto_now=True)

    file_size = models.PositiveIntegerField(default=0)
    name = models.CharField(max_length=100, null=True, blank=True)

    class Meta:
        ordering = ("pk", )

    def save(self, *args, **kwargs):
        if self.media_file and self.mimetype == '':
            # guess mimetype
            mimetype, encoding = mimetypes.guess_type(self.media_file.name)
            if mimetype:
                self.mimetype = mimetype
        if self.media_file and len(self.media_file.name) > 255:
            raise ValueError(
                "Length of the media file should be less or equal to 255")

        try:
            f_size = self.media_file.size
            if f_size:
                self.file_size = f_size
        except (OSError, AttributeError):
            pass

        try:
            self.name = self.filename
            self.extension = self.name.rsplit('.', 1)[1]
        except Exception:
            pass

        super(Attachment, self).save(*args, **kwargs)

    @property
    def file_hash(self):
        if self.media_file.storage.exists(self.media_file.name):
            return u'%s' % md5(self.media_file.read()).hexdigest()
        return u''

    @property
    def filename(self):
        if self.media_file:
            return os.path.basename(self.media_file.name)


def is_newline_error(e):
    """
    Return True is e is a new line error based on the error text.
    Otherwise return False.
    """
    newline_error = u'new-line character seen in unquoted field - do you need'\
        u' to open the file in universal-newline mode?'
    return newline_error == str(e)


def process_xlsform(xls, default_name):
    """
    Process XLSForm file and return the survey dictionary for the XLSForm.
    """
    # FLOW Results package is a JSON file.

    file_object = None
    if xls.name.endswith('csv'):
        # a csv file gets closed in pyxform, make a copy
        xls.seek(0)
        file_object = BytesIO()
        file_object.write(xls.read())
        file_object.seek(0)
        xls.seek(0)

    try:
        return parse_file_to_json(xls.name, file_object=file_object or xls)
    except csv.Error as e:
        if is_newline_error(e):
            xls.seek(0)
            file_object = StringIO(
                u'\n'.join(xls.read().splitlines()))
            return parse_file_to_json(
                xls.name, default_name=default_name, file_object=file_object)
        raise e


def get_columns_with_hxl(survey_elements):
    '''
    Returns a dictionary whose keys are xform field names and values are
    `instance::hxl` values set on the xform
    :param include_hxl - boolean value
    :param survey_elements - survey elements of an xform
    return dictionary or None
    '''
    return survey_elements and {
        se.get('name'): val.get('hxl')
        for se in survey_elements
        for key, val in se.items()
        if key == 'instance' and val and 'hxl' in val
    }


def check_version_set(survey):
    """
    Checks if the version has been set in the xls file and if not adds
    the default version in this datetime (yyyymmddhhmm) format.
    """

    # get the json and check for the version key
    survey_json = json.loads(survey.to_json())
    if not survey_json.get("version"):
        # set utc time as the default version
        survey_json['version'] = \
            timezone.now().strftime("%Y%m%d%H%M")
        builder = SurveyElementBuilder()
        survey = builder.create_survey_element_from_json(
            json.dumps(survey_json))
    return survey


class DataDictionary(XForm):  # pylint: disable=too-many-instance-attributes
    """
    DataDictionary model class.
    """

    def __init__(self, *args, **kwargs):
        self.instances_for_export = lambda d: d.instances.all()
        self.has_external_choices = False
        self._id_string_changed = False
        super(DataDictionary, self).__init__(*args, **kwargs)

    def __str__(self):
        return getattr(self, "id_string", "")

    def save(self, *args, **kwargs):
        skip_xls_read = kwargs.get('skip_xls_read')

        if self.xls and not skip_xls_read:
            default_name = None \
                if not self.pk else self.survey.xml_instance().tagName
            survey_dict = process_xlsform(self.xls, default_name)
            if has_external_choices(survey_dict):
                self.has_external_choices = True
            survey = create_survey_element_from_dict(survey_dict)
            survey = check_version_set(survey)
            if get_columns_with_hxl(survey.get('children')):
                self.has_hxl_support = True
            # if form is being replaced, don't check for id_string uniqueness
            if self.pk is None:
                new_id_string = self.get_unique_id_string(
                    survey.get('id_string'))
                self._id_string_changed = \
                    new_id_string != survey.get('id_string')
                survey['id_string'] = new_id_string
                # For flow results packages use the user defined id/uuid
            elif self.id_string != survey.get('id_string'):
                raise XLSFormError(
                    ("Your updated form's id_string '%(new_id)s' must match "
                     "the existing forms' id_string '%(old_id)s'." % {
                         'new_id': survey.get('id_string'),
                         'old_id': self.id_string}))
            elif default_name and default_name != survey.get('name'):
                survey['name'] = default_name
            else:
                survey['id_string'] = self.id_string
            self.json = survey.to_json()
            self.xml = survey.to_xml()
            self.version = survey.get('version')
            self.last_updated_at = timezone.now()
            self.title = survey.get('title')
            self._mark_start_time_boolean()
            set_uuid(self)
            self._set_uuid_in_xml()
            self._set_hash()

        if 'skip_xls_read' in kwargs:
            del kwargs['skip_xls_read']

        super(DataDictionary, self).save(*args, **kwargs)

    def file_name(self):
        return os.path.split(self.xls.name)[-1]


def sheet_to_csv(xls_content, sheet_name):
    """Writes a csv file of a specified sheet from a an excel file

    :param xls_content: Excel file contents
    :param sheet_name: the name of the excel sheet to generate the csv file

    :returns: a (StrionIO) csv file object
    """
    workbook = xlrd.open_workbook(file_contents=xls_content)

    sheet = workbook.sheet_by_name(sheet_name)

    if not sheet or sheet.nrows < 2:
        raise Exception("Sheet <'%(sheet_name)s'> has no data." % {
            'sheet_name': sheet_name})

    csv_file = BytesIO()

    writer = csv.writer(csv_file, encoding='utf-8', quoting=csv.QUOTE_ALL)
    mask = [v and len(v.strip()) > 0 for v in sheet.row_values(0)]

    header = [v for v, m in zip(sheet.row_values(0), mask) if m]
    writer.writerow(header)

    name_column = None
    try:
        name_column = header.index('name')
    except ValueError:
        pass

    integer_fields = False
    date_fields = False
    if name_column:
        name_column_values = sheet.col_values(name_column)
        for index in range(len(name_column_values)):
            if sheet.cell_type(index, name_column) == xlrd.XL_CELL_NUMBER:
                integer_fields = True
            elif sheet.cell_type(index, name_column) == xlrd.XL_CELL_DATE:
                date_fields = True

    for row in range(1, sheet.nrows):
        if integer_fields or date_fields:
            # convert integers to string/datetime if name has numbers/dates
            row_values = []
            for index, val in enumerate(sheet.row_values(row)):
                if sheet.cell_type(row, index) == xlrd.XL_CELL_NUMBER:
                    try:
                        val = str(
                            float(val) if (
                                float(val) > int(val)
                            ) else int(val)
                        )
                    except ValueError:
                        pass
                elif sheet.cell_type(row, index) == xlrd.XL_CELL_DATE:
                    val = xlrd.xldate_as_datetime(
                        val, workbook.datemode).isoformat()
                row_values.append(val)
            writer.writerow([v for v, m in zip(row_values, mask) if m])
        else:
            writer.writerow(
                [v for v, m in zip(sheet.row_values(row), mask) if m])

    return csv_file


def set_object_permissions(sender, instance=None, created=False, **kwargs):
    """
    Apply the relevant object permissions for the form to all users who should
    have access to it.
    """
    # seems the super is not called, have to get xform from here
    xform = XForm.objects.get(pk=instance.pk)

    if hasattr(instance, 'has_external_choices') \
            and instance.has_external_choices:
        instance.xls.seek(0)
        f = sheet_to_csv(instance.xls.read(), 'external_choices')
        f.seek(0, os.SEEK_END)
        size = f.tell()
        f.seek(0)

        data_file = InMemoryUploadedFile(
            file=f,
            field_name='data_file',
            name='itemsets.csv',
            content_type='text/csv',
            size=size,
            charset=None
        )

        MetaData.media_upload(xform, data_file)


post_save.connect(set_object_permissions, sender=DataDictionary,
                  dispatch_uid='xform_object_permissions')



# Generated by Django 2.2 on 2019-04-29 18:54

from django.conf import settings
import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion
import xform.models


if 'mysql' in settings.DATABASES['default']['ENGINE']:
    from jsonfield import JSONField
else:
    from django.contrib.postgres.fields.jsonb import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='XForm',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('xls', models.FileField(null=True, upload_to=xform.models.upload_to)),
                ('json', models.TextField(default='')),
                ('description', models.TextField(blank=True, default='', null=True)),
                ('xml', models.TextField()),
                ('id_string', models.SlugField(editable=False, max_length=100, verbose_name='ID')),
                ('title', models.CharField(editable=False, max_length=255)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('last_submission_time', models.DateTimeField(blank=True, null=True)),
                ('has_start_time', models.BooleanField(default=False)),
                ('uuid', models.CharField(default='', max_length=36)),
                ('instances_with_geopoints', models.BooleanField(default=False)),
                ('num_of_submissions', models.IntegerField(default=0)),
                ('version', models.CharField(blank=True, max_length=255, null=True)),
                ('has_hxl_support', models.BooleanField(default=False)),
                ('last_updated_at', models.DateTimeField(auto_now=True)),
                ('hash', models.CharField(blank=True, default=None, max_length=36, null=True, verbose_name='Hash')),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='xforms', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'XForm',
                'verbose_name_plural': 'XForms',
                'ordering': ('pk',),
                'unique_together': {('user', 'id_string')},
            },
        ),
        migrations.CreateModel(
            name='MetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data_type', models.CharField(max_length=255)),
                ('data_value', models.CharField(max_length=255)),
                ('data_file', models.FileField(blank=True, null=True, upload_to=xform.models.upload_to)),
                ('data_file_type', models.CharField(blank=True, max_length=255, null=True)),
                ('file_hash', models.CharField(blank=True, max_length=50, null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True)),
                ('date_modified', models.DateTimeField(auto_now=True, null=True)),
                ('deleted_at', models.DateTimeField(default=None, null=True)),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('content_type', models.ForeignKey(default=xform.models.get_default_content_type, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'unique_together': {('object_id', 'data_type', 'data_value', 'content_type')},
            },
        ),
        migrations.CreateModel(
            name='Instance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('json', JSONField(default=dict)),
                ('xml', models.TextField()),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('last_edited', models.DateTimeField(default=None, null=True)),
                ('status', models.CharField(default='submitted_via_web', max_length=20)),
                ('uuid', models.CharField(db_index=True, default='', max_length=249)),
                ('version', models.CharField(max_length=255, null=True)),
                ('geom', django.contrib.gis.db.models.fields.GeometryCollectionField(null=True, srid=4326)),
                ('media_all_received', models.NullBooleanField(default=True, verbose_name='Received All Media Attachemts')),
                ('total_media', models.PositiveIntegerField(default=0, null=True, verbose_name='Total Media Attachments')),
                ('media_count', models.PositiveIntegerField(default=0, null=True, verbose_name='Received Media Attachments')),
                ('checksum', models.CharField(blank=True, db_index=True, max_length=64, null=True)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='instances', to=settings.AUTH_USER_MODEL)),
                ('xform', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='instances', to='xform.XForm')),
            ],
            options={
                'unique_together': {('xform', 'uuid')},
            },
        ),
    ]



# Generated by Django 2.2 on 2019-04-29 19:18

from django.db import migrations, models
import django.db.models.deletion
import xform.models


class Migration(migrations.Migration):

    dependencies = [
        ('xform', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('media_file', models.FileField(max_length=255, upload_to=xform.models.upload_to)),
                ('mimetype', models.CharField(blank=True, default='', max_length=100)),
                ('extension', models.CharField(db_index=True, default='non', max_length=10)),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True)),
                ('date_modified', models.DateTimeField(auto_now=True, null=True)),
                ('file_size', models.PositiveIntegerField(default=0)),
                ('name', models.CharField(blank=True, max_length=100, null=True)),
                ('instance', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='xform.Instance')),
            ],
            options={
                'ordering': ('pk',),
            },
        ),
    ]



import os
import re

from dateutil import parser as pdt
from hashlib import sha256
from xml.parsers.expat import ExpatError

from django.core.exceptions import MultipleObjectsReturned, PermissionDenied
from django.db import IntegrityError, transaction, DataError
from django.db.models import Q
from django.http import UnreadablePostError
from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.utils.encoding import DjangoUnicodeDecodeError

from .models import XForm, Instance, Attachment
from .openrosaresponse import OpenRosaResponse, OpenRosaResponseBadRequest
from .openrosaresponse import OpenRosaResponseForbidden
from .openrosaresponse import OpenRosaResponseNotAllowed
from .openrosaresponse import OpenRosaResponseNotFound
from .utils import get_uuid_from_xml, get_meta_from_xml, clean_and_parse_xml
from .utils import (
    DuplicateInstance, InstanceInvalidUserError, InstanceMultipleNodeError,
    InstanceEmptyError, NonUniqueFormIdError, FormIsMergedDatasetError,
    FormInactiveError
)


uuid_regex = re.compile(r'<formhub>\s*<uuid>\s*([^<]+)\s*</uuid>\s*</formhub>',
                        re.DOTALL)


def get_submission_date_from_xml(xml):
    # check in survey_node attributes
    xml = clean_and_parse_xml(xml)
    children = xml.childNodes
    # children ideally contains a single element
    # that is the parent of all survey elements
    if children.length == 0:
        raise ValueError("XML string must have a survey element.")
    survey_node = children[0]
    submissionDate = survey_node.getAttribute('submissionDate')
    if submissionDate != '':
        return pdt.parse(submissionDate)
    return None


def get_id_string_from_xml_str(xml_str):
    xml_obj = clean_and_parse_xml(xml_str)
    root_node = xml_obj.documentElement
    id_string = root_node.getAttribute(u"id")

    if len(id_string) == 0:
        # may be hidden in submission/data/id_string
        elems = root_node.getElementsByTagName('data')

        for data in elems:
            for child in data.childNodes:
                id_string = data.childNodes[0].getAttribute('id')

                if len(id_string) > 0:
                    break

            if len(id_string) > 0:
                break

    return id_string


def get_uuid_from_submission(xml):
    # parse UUID from uploaded XML
    split_xml = uuid_regex.split(xml.decode('utf-8'))

    # check that xml has UUID
    return len(split_xml) > 1 and split_xml[1] or None


def get_xform_from_submission(xml, username, uuid=None):
    # check alternative form submission ids
    uuid = uuid or get_uuid_from_submission(xml)

    if not username and not uuid:
        raise InstanceInvalidUserError()

    if uuid:
        # try find the form by its uuid which is the ideal condition
        if XForm.objects.filter(
                uuid=uuid).count() > 0:
            xform = XForm.objects.get(uuid=uuid,)

            return xform

    id_string = get_id_string_from_xml_str(xml)

    try:
        return get_object_or_404(
            XForm,
            id_string__iexact=id_string
        )
    except MultipleObjectsReturned:
        raise NonUniqueFormIdError()


def check_edit_submission_permissions(request_user, xform):
    return True


def check_submission_permissions(request, xform):
    # TODO
    return True


def get_first_record(queryset):
    """
    Returns the first item in a queryset sorted by id.
    """
    records = sorted([record for record in queryset], key=lambda k: k.id)
    if records:
        return records[0]

    return None


def get_filtered_instances(*args, **kwargs):
    """Get filtered instances - mainly to allow mocking in tests"""

    return Instance.objects.filter(*args, **kwargs)


def update_attachment_tracking(instance):
    """
    Takes an Instance object and updates attachment tracking fields
    """
    instance.skip_signal = True
    instance.total_media = instance.num_of_media
    instance.media_count = instance.attachments_count
    instance.media_all_received = instance.media_count == instance.total_media
    instance.save(update_fields=['total_media', 'media_count',
                                 'media_all_received', 'json'])


def save_attachments(xform, instance, media_files):
    """
    Saves attachments for the given instance/submission.
    """
    # upload_path = os.path.join(instance.xform.user.username, 'attachments')

    for f in media_files:
        filename, extension = os.path.splitext(f.name)
        extension = extension.replace('.', '')
        content_type = u'text/xml' \
            if extension == Attachment.OSM else f.content_type
        if extension == Attachment.OSM and not xform.instances_with_osm:
            xform.instances_with_osm = True
            xform.save()
        filename = os.path.basename(f.name)
        media_in_submission = (
            filename in instance.get_expected_media() or
            instance.xml.decode('utf-8').find(filename) != -1)
        if media_in_submission:
            Attachment.objects.get_or_create(
                instance=instance,
                media_file=f,
                mimetype=content_type,
                name=filename,
                extension=extension)
    update_attachment_tracking(instance)


def get_deprecated_uuid_from_xml(xml):
    uuid = get_meta_from_xml(xml, "deprecatedID")
    regex = re.compile(r"uuid:(.*)")
    if uuid:
        matches = regex.match(uuid)
        if matches and len(matches.groups()) > 0:
            return matches.groups()[0]
    return None


def _get_instance(xml, new_uuid, submitted_by, status, xform, checksum):
    history = None
    instance = None

    if isinstance(xml, bytes):
        xml = xml.decode('utf-8')

    # check if its an edit submission
    old_uuid = get_deprecated_uuid_from_xml(xml)
    if old_uuid:
        instance = Instance.objects.filter(uuid=old_uuid,
                                           xform_id=xform.pk).first()
        if instance:
            # edits
            last_edited = timezone.now()
            instance.xml = xml
            instance.last_edited = last_edited
            instance.uuid = new_uuid
            instance.checksum = checksum
            instance.save()

    if old_uuid is None or (instance is None and history is None):
        # new submission
        instance = Instance(
            xml=xml, user=submitted_by, status=status, xform=xform,
            checksum=checksum)
        instance.skip_signal = False
        instance.save()
    return instance


def save_submission(xform, xml, media_files, new_uuid, submitted_by, status, checksum):
    date_created_override = get_submission_date_from_xml(xml)
    instance = _get_instance(xml, new_uuid, submitted_by, status, xform, checksum)
    save_attachments(xform, instance, media_files)

    # override date created if required
    if date_created_override:
        if not timezone.is_aware(date_created_override):
            # default to utc?
            date_created_override = timezone.make_aware(
                date_created_override, timezone.utc)
        instance.date_created = date_created_override
        instance.save()

    if instance.xform is not None:
        instance.skip_signal = True
        instance.save()
    return instance


def create_instance(
    username,
    xml_file,
    media_files,
    status=u'submitted_via_web',
    uuid=None,
    request=None
):
    """
    I used to check if this file had been submitted already, I've
    taken this out because it was too slow. Now we're going to create
    a way for an admin to mark duplicate instances. This should
    simplify things a bit.
    Submission cases:
    * If there is a username and no uuid, submitting an old ODK form.
    * If there is a username and a uuid, submitting a new ODK form.
    """
    instance = None
    submitted_by = request.user \
        if request and request.user.is_authenticated else None

    if username:
        username = username.lower()

    xml = xml_file.read()
    xform = get_xform_from_submission(xml, username, uuid)
    check_submission_permissions(request, xform)
    checksum = sha256(xml).hexdigest()

    new_uuid = get_uuid_from_xml(xml)
    filtered_instances = get_filtered_instances(
        Q(checksum=checksum) | Q(uuid=new_uuid), xform_id=xform.pk)
    existing_instance = get_first_record(filtered_instances.only('id'))
    if existing_instance and \
            (new_uuid or existing_instance.xform.has_start_time):
        with transaction.atomic():
            save_attachments(xform, existing_instance, media_files)
            existing_instance.save(update_fields=['json', 'date_modified'])
        return DuplicateInstance()

    try:
        with transaction.atomic():
            instance = save_submission(xform, xml, media_files, new_uuid,
                                       submitted_by, status, checksum)
    except IntegrityError:
        instance = get_first_record(Instance.objects.filter(
            Q(checksum=checksum) | Q(uuid=new_uuid),
            xform_id=xform.pk))

        if instance:
            attachment_names = [
                a.media_file.name.split('/')[-1]
                for a in Attachment.objects.filter(instance=instance)
            ]
            media_files = [f for f in media_files
                           if f.name not in attachment_names]
            save_attachments(xform, instance, media_files)
            instance.save()

        instance = DuplicateInstance()
    return instance


def safe_create_instance(username, xml_file, media_files, uuid, request):
    error = instance = None

    try:
        instance = create_instance(
            username, xml_file, media_files, uuid=uuid, request=request)
    except InstanceInvalidUserError:
        error = OpenRosaResponseBadRequest("Username or ID required.")
    except InstanceEmptyError:
        error = OpenRosaResponseBadRequest(
            "Received empty submission. No instance was created")
    except (FormInactiveError, FormIsMergedDatasetError) as e:
        error = OpenRosaResponseNotAllowed(str(e))
    except XForm.DoesNotExist:
        error = OpenRosaResponseNotFound(
            "Form does not exist on this account")
    except ExpatError:
        error = OpenRosaResponseBadRequest("Improperly formatted XML.")
    except DuplicateInstance:
        response = OpenRosaResponse("Duplicate submission")
        response.status_code = 202
        if request:
            response['Location'] = request.build_absolute_uri(request.path)
        error = response
    except PermissionDenied as e:
        error = OpenRosaResponseForbidden(e)
    except UnreadablePostError as e:
        error = OpenRosaResponseBadRequest(
            "Unable to read submitted file: %(error)s" % {'error': str(e)})
    except InstanceMultipleNodeError as e:
        error = OpenRosaResponseBadRequest(e)
    except DjangoUnicodeDecodeError:
        error = OpenRosaResponseBadRequest(
            "File likely corrupted during "
            "transmission, please try later.")
    except NonUniqueFormIdError as e:
        error = OpenRosaResponseBadRequest(
            "Unable to submit because there are multiple forms with"
            " this formID.")
    except DataError as e:
        error = OpenRosaResponseBadRequest((str(e)))
    if isinstance(instance, DuplicateInstance):
        response = OpenRosaResponse("Duplicate submission")
        response.status_code = 202
        if request:
            response['Location'] = request.build_absolute_uri(request.path)
        error = response
        instance = None
    return [error, instance]



from django.http import HttpResponse
from django.utils import timezone

OPEN_ROSA_VERSION_HEADER = 'X-OpenRosa-Version'
OPEN_ROSA_VERSION = '1.0'
DEFAULT_CONTENT_TYPE = 'text/xml; charset=utf-8'


class BaseOpenRosaResponse(HttpResponse):
    status_code = 201

    def __init__(self, *args, **kwargs):
        super(BaseOpenRosaResponse, self).__init__(*args, **kwargs)

        self[OPEN_ROSA_VERSION_HEADER] = OPEN_ROSA_VERSION

        dt = timezone.now().strftime('%a, %d %b %Y %H:%M:%S %Z')
        self['Date'] = dt
        self['X-OpenRosa-Accept-Content-Length'] = 10000000
        self['Content-Type'] = DEFAULT_CONTENT_TYPE


class OpenRosaResponse(BaseOpenRosaResponse):
    status_code = 201

    def __init__(self, *args, **kwargs):
        super(OpenRosaResponse, self).__init__(*args, **kwargs)
        self.message = self.content
        # wrap content around xml
        self.content = '''<?xml version='1.0' encoding='UTF-8' ?>
<OpenRosaResponse xmlns="http://openrosa.org/http/response">
        <message nature="">%s</message>
</OpenRosaResponse>''' % self.content


class OpenRosaResponseNotFound(OpenRosaResponse):
    status_code = 404


class OpenRosaResponseBadRequest(OpenRosaResponse):
    status_code = 400


class OpenRosaResponseNotAllowed(OpenRosaResponse):
    status_code = 405


class OpenRosaResponseForbidden(OpenRosaResponse):
    status_code = 403



# Generated by Django 2.2 on 2019-04-30 21:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('xform', '0002_attachment'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataDictionary',
            fields=[
                ('xform_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='xform.XForm')),
            ],
            bases=('xform.xform',),
        ),
    ]



from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404


from .models import Instance
from .utils import (
    enketo_url, EnketoError, get_form_url, inject_instanceid,
)


def enter_data(request, id_string, offline=False):
    form_url = get_form_url(request, settings.ENKETO_PROTOCOL)
    url = enketo_url(form_url, id_string, offline=offline)
    return HttpResponseRedirect(url)


def edit_data(request, id_string, data_id):
    """
    Redirects to Enketo webform to edit a submission with the data_id.
    """
    return_url_get = request.GET.get('return_url_xform')
    instance = get_object_or_404(
        Instance.objects.only('xml', 'uuid'),
        pk=data_id,
        xform__id_string=id_string
    )

    url = '%sdata/edit_url' % settings.ENKETO_URL
    injected_xml = inject_instanceid(instance.xml, instance.uuid)
    return_url = request.build_absolute_uri(return_url_get)
    form_url = get_form_url(request, settings.ENKETO_PROTOCOL)

    try:
        url = enketo_url(
            form_url,
            id_string,
            instance_xml=injected_xml,
            instance_id=instance.uuid,
            return_url=return_url)
    except EnketoError as e:
        if settings.SENTRY_DSN:
            import sentry_sdk
            sentry_sdk.capture_exception(e)
        messages.error(
            request,
            "Enketo error: enketo replied - %s" % e
        )
    else:
        if url:
            return HttpResponseRedirect(url)
    return HttpResponseRedirect(return_url_get)



import json
import importlib

# from pyxform.xls2json import parse_file_to_json
from pyxform.builder import create_survey_element_from_dict

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import Http404, HttpResponseBadRequest, HttpResponse
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.http import UnreadablePostError
from django.shortcuts import get_object_or_404
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.cache import never_cache

from rest_framework import permissions
from rest_framework import status
from rest_framework import viewsets
from rest_framework.authentication import BasicAuthentication
from rest_framework.authentication import SessionAuthentication
from rest_framework.authentication import get_authorization_header
from rest_framework.authentication import TokenAuthentication
from rest_framework.decorators import action
from rest_framework.generics import CreateAPIView
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.settings import api_settings

# from .authentication import DigestAuthentication
from .forms import QuickConverter
from .models import XForm, MetaData, check_version_set
from .openrosaresponse import OpenRosaResponseBadRequest
from .renderers import MediaFileContentNegotiation
from .renderers import XFormManifestRenderer, XFormListRenderer
from .serializers import XFormListSerializer, SubmissionSerializer, XFormManifestSerializer
from .tags import GROUP_DELIMETER_TAG, REPEAT_INDEX_TAGS
from .utils import get_media_file_response, publish_form


Usuario = get_user_model()


class CsrfExemptSessionAuthentication(SessionAuthentication):

    def enforce_csrf(self, request):
        return  # To not perform the csrf check previously happening


def get_from_module(module_name, function_name):
    module = importlib.import_module(module_name)
    return getattr(module, function_name)


class IsAuthenticatedSubmission(BasePermission):
    """
    IsAuthenticatedSubmission - checks if profile requires authentication
    during a submission request.
    """

    def has_permission(self, request, view):
        username = view.kwargs.get('username')  # noqa
        if request.method in ['HEAD', 'POST'] and request.user.is_anonymous:
            return False

        return True


def get_forms_shared_with_user(user):
    """
    Return forms shared with a user
    """
    xforms = XForm.objects.filter()

    return xforms.exclude(user=user).select_related('user')


class EnketoODKAuthMixin(object):
    def get_authenticators(self):
        try:
            # check if you are the User-Agent of ODK
            # https://github.com/opendatakit/collect/blob/81b105cef60a113efd1954d782648219ec4733e6/collect_app/src/main/java/org/odk/collect/android/application/Collect.java#L221
            if 'org.odk.collect.android' in self.request.META['HTTP_USER_AGENT']:
                return [BasicAuthentication()]
        except Exception:
            pass
        return [CsrfExemptSessionAuthentication()]


class XLSFormError(Exception):
    pass


class AuthenticateHeaderMixin(object):
    def get_authenticate_header(self, request):
        auth = get_authorization_header(request).split()

        if auth and auth[0].lower() == b'token':
            return TokenAuthentication().authenticate_header(request)

        return super(AuthenticateHeaderMixin, self)\
            .get_authenticate_header(request)


def get_openrosa_headers(request, location=True):
    """
    Returns a dict with OpenRosa headers 'Date', 'X-OpenRosa-Version',
    'X-OpenRosa-Accept-Content-Length' and 'Location'.
    """
    now = timezone.now()
    data = {
        'Date': now.strftime('%a, %d %b %Y %H:%M:%S %Z'),
        'X-OpenRosa-Version': '1.0',
        'X-OpenRosa-Accept-Content-Length': 10000000
    }

    if location:
        data['Location'] = request.build_absolute_uri(request.path).replace('http', 'https')

    return data


class OpenRosaHeadersMixin(object):  # pylint: disable=R0903
    """
    OpenRosaHeadersMixin class - sets OpenRosa headers in a response for a View
    or Viewset.
    """

    def finalize_response(self, request, response, *args, **kwargs):
        """
        Adds OpenRosa headers into the response.
        """
        self.headers.update(get_openrosa_headers(request))

        return super(OpenRosaHeadersMixin, self).finalize_response(
            request, response, *args, **kwargs)


class CreateModelMixin(object):
    """
    Create a model instance.
    """

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        serializer.save()

    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}


class XFormListViewSet(EnketoODKAuthMixin, viewsets.ReadOnlyModelViewSet):
    template_name = 'xform/xformsList.xml'
    serializer_class = XFormListSerializer
    permission_classes = (permissions.AllowAny,)
    renderer_classes = (XFormListRenderer,)
    content_negotiation_class = MediaFileContentNegotiation
    queryset = XForm.objects.filter().only(*getattr(
        settings, 'XFORM_ONLY_FILTER', [
            'id_string', 'title', 'version', 'uuid',
            'description', 'hash'
        ]
    ))

    def filter_queryset(self, qs):
        xform_id = self.request.GET.get('formID')
        if xform_id:
            qs = qs.filter(id_string=xform_id)

        qs_filter = getattr(settings, 'XFORM_QS_FILTER', lambda qs, user: qs)

        if isinstance(qs_filter, str):
            module_name, function_name = qs_filter.rsplit(".", 1)
            qs_filter = get_from_module(module_name, function_name)
        return qs_filter(qs, self.request)

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        filter_kwargs = {self.lookup_field: self.kwargs[self.lookup_field]}
        obj = get_object_or_404(queryset or XForm, **filter_kwargs)
        return obj

    def get_renderers(self):
        if self.action and self.action == 'manifest':
            return [XFormManifestRenderer()]

        return super(XFormListViewSet, self).get_renderers()

    @never_cache
    def list(self, request, *args, **kwargs):
        if self.request.user.is_anonymous:
            self.permission_denied(self.request)

        self.object_list = self.filter_queryset(self.get_queryset())

        headers = get_openrosa_headers(request, location=False)
        serializer = self.get_serializer(self.object_list, many=True)
        if request.method in ['HEAD']:
            return Response('', headers=headers, status=204)

        return Response(serializer.data, headers=headers)

    def retrieve(self, request, *args, **kwargs):
        headers = get_openrosa_headers(request, location=False)
        self.object = self.get_object()
        if self.object.dynamic_choices:
            new_choices = {}
            # survey_dict = parse_file_to_json(
            #     json.loads(self.object.json)['name'],
            #     file_object=self.object.xls
            # )
            survey_dict = json.loads(self.object.json)
            fields_keys = settings.XFORM_DYNAMIC_CHOICES.keys()

            def fill_dynamic_choices(els):
                for field in els:
                    if field['name'] in fields_keys:
                        module_name, function_name = settings.XFORM_DYNAMIC_CHOICES[
                            field['name']
                        ].rsplit(".", 1)
                        f = get_from_module(module_name, function_name)
                        field['children'] = f(request, self.object)
                        new_choices[field['name']] = field['children']

                        if 'choices' in field:
                            field['choices'] = []
                            del field['choices']
                    if field['type'] == 'repeat':
                        fill_dynamic_choices(field['children'])

            fill_dynamic_choices(survey_dict['children'])

            if 'choices' in survey_dict:
                survey_dict['choices'].update(new_choices)
            else:
                survey_dict['choices'] = new_choices

            survey = create_survey_element_from_dict(survey_dict)
            survey = check_version_set(survey)

            return Response(
                survey.to_xml(),
                headers=headers
            )

        return Response(self.object.xml,
                        headers=headers)

    @action(methods=['GET', 'HEAD'], detail=True)
    def manifest(self, request, *args, **kwargs):
        self.object = self.get_object()
        object_list = MetaData.objects.filter(object_id=self.object.pk).filter(
            data_type__in=['media', 'url']
        )
        context = self.get_serializer_context()
        context[GROUP_DELIMETER_TAG] = '.'
        context[REPEAT_INDEX_TAGS] = '_,_'
        serializer = XFormManifestSerializer(object_list, many=True,
                                             context=context)
        return Response(serializer.data,
                        headers=get_openrosa_headers(request, location=False))

    @action(methods=['GET', 'HEAD'], detail=True)
    def media(self, request, *args, **kwargs):
        self.object = self.get_object()
        pk = kwargs.get('metadata')

        if not pk:
            raise Http404()

        meta_obj = get_object_or_404(
            MetaData, data_type__in=['media', 'url'],
            object_id=self.object.pk, pk=pk)
        response = get_media_file_response(
            meta_obj,
            username=self.kwargs['username']
        )

        if response.status_code == 403 and request.user.is_anonymous:
            # raises a permission denied exception, forces authentication
            self.permission_denied(request)
        else:
            return response


class XFormSubmissionView(EnketoODKAuthMixin, OpenRosaHeadersMixin, CreateAPIView):
    '''In the inheritance being a CrateView, but this view is also used for updates.
        The serializer's create method is always called.
        The create method calls a util named create_instance that checks for a
        <deprecatedID> in xml, if there is this tag it will retrieve the instance to be updated.
    '''

    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = SubmissionSerializer
    template_name = 'xform/submission.xml'

    def head(self, request, format=None):
        resp = HttpResponse(b'', status=204)
        return resp

    def handle_exception(self, exc):
        """
        Handles exceptions thrown by handler method and
        returns appropriate error response.
        """
        if hasattr(exc, 'response'):
            return exc.response

        if isinstance(exc, UnreadablePostError):
            return OpenRosaResponseBadRequest(
                "Unable to read submitted file, please try re-submitting.")

        return super(XFormSubmissionView, self).handle_exception(exc)


def profile(request, username):
    content_user = get_object_or_404(Usuario, cpf__iexact=username)
    form = QuickConverter()
    data = {'form': form}

    # xlsform submission...
    if request.method == 'POST' and request.user.is_authenticated:
        def set_form():
            form = QuickConverter(request.POST, request.FILES)
            form.publish(request.user)
        form_result = publish_form(set_form)
        data['message'] = form_result

    if content_user == request.user:
        data.update({
            'form': form,
        })
    try:
        resp = render(request, "xform/profile.html", data)
    except XLSFormError as e:
        resp = HttpResponseBadRequest(e.__str__())

    return resp


def edit(request, username, id_string):
    if not request.method == 'POST':
        return HttpResponseForbidden('Update failed.')

    xform = XForm.objects.get(id_string__iexact=id_string)
    if username == request.user.username or\
            request.user.has_perm('logger.change_xform', xform):

        if request.FILES.get('media'):
            for aFile in request.FILES.getlist("media"):
                MetaData.media_upload(xform, aFile)
        if request.POST.get('media_url'):
            for url in request.POST.getlist("media_url"):
                MetaData.add_url(xform, url)
        xform.save()
        return HttpResponseRedirect(request.GET.get('next_url'))
    return HttpResponseForbidden('Update failed.')



from io import BytesIO

from django.utils.xmlutils import SimplerXMLGenerator

from rest_framework import negotiation
from rest_framework.renderers import BaseRenderer, TemplateHTMLRenderer
from rest_framework_xml.renderers import XMLRenderer


FORMLIST_MANDATORY_FIELDS = [
    'formID',
    'name',
    'version',
    'hash',
    'downloadUrl'
]


class XFormListRenderer(BaseRenderer):
    """
    Renderer which serializes to XML.
    """

    media_type = 'text/xml'
    format = 'xml'
    charset = 'utf-8'
    root_node = 'xforms'
    element_node = 'xform'
    xmlns = "http://openrosa.org/xforms/xformsList"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Renders *obj* into serialized XML.
        """
        if data is None:
            return ''
        elif isinstance(data, str):
            return data

        stream = BytesIO()

        xml = SimplerXMLGenerator(stream, self.charset)
        xml.startDocument()
        xml.startElement(self.root_node, {'xmlns': self.xmlns})

        self._to_xml(xml, data)

        xml.endElement(self.root_node)
        xml.endDocument()

        return stream.getvalue()

    def _to_xml(self, xml, data):
        if isinstance(data, (list, tuple)):
            for item in data:
                xml.startElement(self.element_node, {})
                self._to_xml(xml, item)
                xml.endElement(self.element_node)

        elif isinstance(data, dict):
            for (key, value) in iter(data.items()):
                if key not in FORMLIST_MANDATORY_FIELDS and value is None:
                    continue
                xml.startElement(key, {})
                self._to_xml(xml, value)
                xml.endElement(key)

        elif data is None:
            # Don't output any value
            pass

        else:
            xml.characters(str(data))


class XFormManifestRenderer(XFormListRenderer):  # pylint: disable=R0903
    """
    XFormManifestRenderer - render XFormManifest XML.
    """
    root_node = "manifest"
    element_node = "mediaFile"
    xmlns = "http://openrosa.org/xforms/xformsManifest"


class TemplateXMLRenderer(TemplateHTMLRenderer):  # pylint: disable=R0903
    """
    TemplateXMLRenderer - Render XML template.
    """
    format = 'xml'
    media_type = 'text/xml'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        renderer_context = renderer_context or {}
        response = renderer_context['response']

        if response and response.exception:
            return XMLRenderer().render(data, accepted_media_type,
                                        renderer_context)

        return super(TemplateXMLRenderer,
                     self).render(data, accepted_media_type, renderer_context)


class MediaFileContentNegotiation(negotiation.DefaultContentNegotiation):
    """
    MediaFileContentNegotiation - filters renders to only return renders with
                                  matching format.
    """

    def filter_renderers(self, renderers, format):  # pylint: disable=W0622
        """
        If there is a '.json' style format suffix, filter the renderers
        so that we only negotiation against those that accept that format.
        If there is no renderer available, we use MediaFileRenderer.
        """
        renderers = [
            renderer for renderer in renderers if renderer.format == format
        ]
        if not renderers:
            renderers = [MediaFileRenderer()]

        return renderers


class MediaFileRenderer(BaseRenderer):  # pylint: disable=R0903
    """
    MediaFileRenderer - render binary media files.
    """
    media_type = '*/*'
    format = None
    charset = None
    render_style = 'binary'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        return data



import mimetypes

from functools import wraps

from django.conf import settings

from rest_framework import exceptions
from rest_framework import serializers
from rest_framework.reverse import reverse

from .models import MetaData
from .safe_create_instance import safe_create_instance
from .tags import GROUP_DELIMETER_TAG, REPEAT_INDEX_TAGS
from .utils import get_file_extension


def get_request_and_username(context):
    """
    Returns request object and username
    """
    request = context['request']
    view = context['view']
    username = view.kwargs.get('username')

    if not username:
        # get the username from the user if not set
        username = (request.user and request.user.username)

    return (request, username)


def check_obj(f):
    @wraps(f)
    def with_check_obj(*args, **kwargs):
        if args[0]:
            return f(*args, **kwargs)

    return with_check_obj


class XFormListSerializer(serializers.Serializer):
    formID = serializers.ReadOnlyField(source='id_string')
    name = serializers.ReadOnlyField(source='title')
    version = serializers.SerializerMethodField()
    hash = serializers.ReadOnlyField(source='random_hash' if settings.XFORM_RANDOM_HASH else 'hash')
    descriptionText = serializers.ReadOnlyField(source='description')
    downloadUrl = serializers.SerializerMethodField('get_url')
    manifestUrl = serializers.SerializerMethodField('get_manifest_url')

    @check_obj
    def get_version(self, obj):
        if obj.version and obj.version.isdigit():
            return obj.version

    @check_obj
    def get_url(self, obj):
        request = self.context.get('request')
        kwargs = {'pk': obj.pk, 'username': request.user.username}
        return reverse('download_xform', kwargs=kwargs, request=request)

    @check_obj
    def get_manifest_url(self, obj):
        request = self.context.get('request')
        kwargs = {'pk': obj.pk, 'username': request.user.username}
        object_list = MetaData.objects.filter(data_type__in=('media', 'url'),
                                              object_id=obj.pk)
        if object_list:
            return reverse('manifest-url', kwargs=kwargs, request=request)
        return None


class XFormManifestSerializer(serializers.Serializer):
    filename = serializers.SerializerMethodField()
    hash = serializers.SerializerMethodField()
    downloadUrl = serializers.SerializerMethodField('get_url')

    @check_obj
    def get_url(self, obj):
        request = self.context.get('request')
        username = self.context['view'].kwargs['username']
        kwargs = {
            'pk': obj.content_object.pk,
            'username': username,
            'metadata': obj.pk
        }
        try:
            fmt = obj.data_value[obj.data_value.rindex('.') + 1:]
            if obj.data_type == 'url':
                ct, enc = mimetypes.guess_type(fmt)
                if not ct:
                    file_data = obj.file(username=username)
                    fmt = get_file_extension(file_data.content_type)
        except ValueError:
            fmt = 'csv'
        url = reverse(
            'xform-media', kwargs=kwargs, request=request, format=fmt.lower())

        group_delimiter = self.context.get(GROUP_DELIMETER_TAG)
        repeat_index_tags = self.context.get(REPEAT_INDEX_TAGS)
        if group_delimiter and repeat_index_tags and fmt == 'csv':
            url = (url + "?%s=%s&%s=%s" % (
                GROUP_DELIMETER_TAG, group_delimiter, REPEAT_INDEX_TAGS,
                repeat_index_tags))
        return url

    @check_obj
    def get_hash(self, obj):
        if obj.data_type == 'url':
            file_data = obj.file(username=self.context['view'].kwargs['username'])
            hmd5 = MetaData.get_md5(file_data)
            return hmd5
        return obj.file_hash

    @check_obj
    def get_filename(self, obj):
        if obj.data_type == 'url':
            file_data = obj.file(username=self.context['view'].kwargs['username'])
            return file_data.name
        return obj.data_value


class SubmissionSuccessMixin(object):  # pylint: disable=R0903
    """
    SubmissionSuccessMixin - prepares submission success data/message.
    """

    def to_representation(self, instance):
        """
        Returns a dict with a successful submission message.
        """
        if instance is None:
            return super(SubmissionSuccessMixin, self)\
                .to_representation(instance)

        return {
            'message': "Recebido com sucesso.",
            'formid': instance.xform.id_string,
            'encrypted': False,
            'instanceID': u'uuid:%s' % instance.uuid,
            'submissionDate': instance.date_created.isoformat(),
            'markedAsCompleteDate': instance.date_modified.isoformat()
        }


class SubmissionSerializer(SubmissionSuccessMixin, serializers.Serializer):
    """
    XML SubmissionSerializer - handles creating a submission from XML.
    """

    def validate(self, attrs):
        request, __ = get_request_and_username(self.context)
        if not request.FILES or 'xml_submission_file' not in request.FILES:
            raise serializers.ValidationError("No XML submission file.")

        return super(SubmissionSerializer, self).validate(attrs)

    def create(self, validated_data):
        """
        Returns object instances based on the validated data
        """
        request, username = get_request_and_username(self.context)

        xml_file_list = request.FILES.pop('xml_submission_file', [])
        xml_file = xml_file_list[0] if xml_file_list else None
        media_files = request.FILES.values()

        error, instance = safe_create_instance(username, xml_file, media_files,
                                               None, request)
        if error:
            exc = exceptions.APIException(detail=error)
            exc.response = error
            exc.status_code = error.status_code

            raise exc

        return instance



import os
import re

from setuptools import find_packages, setup


def read(f):
    return open(f, 'r', encoding='utf-8').read()


def get_version(package):
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setup(
    name='djangoxform',
    version=get_version('xform'),
    url='https://github.com/znc-sistemas/Django-XForm',
    license='MIT',
    description='OpenRosa for Django.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='NECTO',
    author_email='contato@nectosystems.com.br',  # SEE NOTE BELOW (*)
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=[
        'requests==2.21.0',
        'pyxform==0.13.1',
        'xlrd==1.2.0',
        'djangorestframework==3.9.2',
        'djangorestframework-xml==1.4.0',
    ],
    python_requires=">=3.5",
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



# Generated by Django 2.2 on 2019-05-06 23:01

from django.db import migrations, models
import xform.models


class Migration(migrations.Migration):

    dependencies = [
        ('xform', '0003_datadictionary'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metadata',
            name='data_file',
            field=models.FileField(blank=True, null=True, upload_to=xform.models.meta_data_upload_to),
        ),
    ]
